@@ -17,6 +17,7 @@ from requests import HTTPError
 API_AUTHORITY = "https://apis.fedex.com/"
 AUTH_PATH = "oauth/token"
 TRACK_PATH = "track/v1/trackingnumbers"
+MULTIPIECE_PATH = "track/v1/associatedshipments"
 TIMEOUT = 10
 
 
@@ -109,6 +110,15 @@ class TrackingResult:
     latest_status: DateAndTimeEvent = None
     events: list = None
     package: Package = None
+    json: dict = None
+    has_associated_shipments: bool = False
+    master_tracking_number: str = None
+
+
+@dataclass
+class MultipieceShipment:
+    tracking_number: str
+    tracking_results: list
 
 
 def _download_file(url, filename):
@@ -120,14 +130,74 @@ def _download_file(url, filename):
     return filename
 
 
+def process_track_results(json):
+    tracking_results = []
+    for track_index in json["output"]["completeTrackResults"]:
+        track_results = track_index["trackResults"]
+        for tracking_result in track_results:
+            tracking_number_info = tracking_result["trackingNumberInfo"]
+            tracking_number = tracking_number_info["trackingNumber"]
+            unique_id = tracking_number_info["trackingNumberUniqueId"]
+            carrier_code = tracking_number_info["carrierCode"]
+            shipped = False
+            delivered = False
+            ship_date = None  # declare out of the loop scope to preserve
+            delivery_date = None
+            latest_date = None
+            latest_type = None
+            events = []
+            if "dateAndTimes" not in tracking_result:
+                return TrackingResult(False)
+            for event in tracking_result["dateAndTimes"]:
+                event_type = Event[event["type"]]
+                event_date = datetime.fromisoformat(event["dateTime"])
+                if event_type == Event.SHIP:
+                    shipped = True
+                    ship_date = event_date
+                elif event_type == Event.ACTUAL_DELIVERY:
+                    delivered = True
+                    delivery_date = event_date
+                elif event_type == Event.ESTIMATED_DELIVERY:
+                    delivery_date = event_date
+                if latest_type is None or event_date > latest_date:
+                    latest_date = event_date
+                    latest_type = event_type
+                new_event = DateAndTimeEvent(event_date, event_type)
+                events.append(new_event)
+            latest_event = DateAndTimeEvent(latest_date, latest_type)
+            # package = Package(PackageType[tracking_result["packageDetails"]["physicalPackagingType"]],
+            #                   int(tracking_result["packageDetails"]["count"]))
+            has_associated_shipments = tracking_result["additionalTrackingInfo"]["hasAssociatedShipments"]
+            master_tracking_number = ""
+            if has_associated_shipments:
+                master_tracking_number = tracking_result["additionalTrackingInfo"]["packageIdentifiers"][0]["values"][0]
+            new_tracking_result = TrackingResult(
+                is_valid=True,
+                tracking_number=tracking_number,
+                unique_id=unique_id,
+                carrier_code=carrier_code,
+                is_delivered=delivered,
+                is_shipped=shipped,
+                date_ship=ship_date,
+                date_delivery=delivery_date,
+                latest_status=latest_event,
+                events=events,
+                json=json,
+                has_associated_shipments=has_associated_shipments,
+                master_tracking_number=master_tracking_number)
+            tracking_results.append(new_tracking_result)
+    return tracking_results
+
 class FedexAPI:
 
-    def __init__(self, api_key, secret_key, api_authority=API_AUTHORITY, auth_path=AUTH_PATH, track_path=TRACK_PATH):
+    def __init__(self, api_key, secret_key, api_authority=API_AUTHORITY, auth_path=AUTH_PATH, track_path=TRACK_PATH,
+                 multipiece_path=MULTIPIECE_PATH):
         if secret_key == "" or api_key == "":
             raise AuthenticationError("\nInvalid key supplied")
         self.API_URL = api_authority
         self.AUTH_URL = api_authority + auth_path
         self.TRACK_URL = api_authority + track_path
+        self.MULTIPIECE_URL = api_authority + multipiece_path
         self.API_KEY = api_key
         self.SECRET_KEY = secret_key
         request = requests.post(self.AUTH_URL,
@@ -157,63 +227,41 @@ class FedexAPI:
         response_data = response.json()
         if "errors" in response_data:
             return TrackingResult(False)
-        tracking_results = []
-        for track_index in response_data["output"]["completeTrackResults"]:
-            track_results = track_index["trackResults"]
-            for tracking_result in track_results:
-                tracking_number_info = tracking_result["trackingNumberInfo"]
-                tracking_number = tracking_number_info["trackingNumber"]
-                unique_id = tracking_number_info["trackingNumberUniqueId"]
-                carrier_code = tracking_number_info["carrierCode"]
-                shipped = False
-                delivered = False
-                ship_date = None  # declare out of the loop scope to preserve
-                delivery_date = None  # declare out of the loop scope to preserve
-                latest_date = None  # declare out of the loop scope to preserve
-                latest_type = None  # declare out of the loop scope to preserve
-                events = []
-                if "dateAndTimes" not in tracking_result:
-                    return TrackingResult(False)
-                for event in tracking_result["dateAndTimes"]:
-                    event_type = Event[event["type"]]
-                    event_date = datetime.fromisoformat(event["dateTime"])
-                    if event_type == Event.SHIP:
-                        shipped = True
-                        ship_date = event_date
-                    elif event_type == Event.ACTUAL_DELIVERY:
-                        delivered = True
-                        delivery_date = event_date
-                    elif event_type == Event.ESTIMATED_DELIVERY:
-                        delivery_date = event_date
-                    if latest_type is None or event_date > latest_date:
-                        latest_date = event_date
-                        latest_type = event_type
-                    new_event = DateAndTimeEvent(event_date, event_type)
-                    events.append(new_event)
-                latest_event = DateAndTimeEvent(latest_date, latest_type)
-                # package = Package(PackageType[tracking_result["packageDetails"]["physicalPackagingType"]],
-                #                   int(tracking_result["packageDetails"]["count"]))
-                new_tracking_result = TrackingResult(
-                    is_valid=True,
-                    tracking_number=tracking_number,
-                    unique_id=unique_id,
-                    carrier_code=carrier_code,
-                    is_delivered=delivered,
-                    is_shipped=shipped,
-                    date_ship=ship_date,
-                    date_delivery=delivery_date,
-                    latest_status=latest_event,
-                    events=events)
-                tracking_results.append(new_tracking_result)
-        # prevent duplicate errors
+        tracking_results = process_track_results(response_data)
         latest_date = None  # declare out of the loop scope to preserve
-        latest_result = None  # declare out of the loop scope to preserve
+        latest_result = None
         for result in tracking_results:
             if latest_result is None or result.latest_status.datetime > latest_date:
                 latest_date = result.latest_status.datetime
                 latest_result = result
         return latest_result
 
+    def track_multipiece(self, tracking_number):
+        payload = {
+            "masterTrackingNumberInfo": {
+                "trackingNumberInfo": {
+                    "trackingNumber": str(tracking_number)
+                }
+            },
+            "associatedType": "STANDARD_MPS"
+        }
+        headers = {"Content-type": "application/json",
+                   "Authorization": f"Bearer {self.AUTH_KEY}"}
+        response = requests.request("POST", self.MULTIPIECE_URL,
+                                    data=str(payload).replace("'", '"'),
+                                    headers=headers)
+        try:
+            response.raise_for_status()
+        except HTTPError:
+            return TrackingResult(False)
+        response_data = response.json()
+        tracking_results = process_track_results(response_data)
+        tracking_number = tracking_results[0].tracking_number
+        mps = MultipieceShipment(str(tracking_number), tracking_results)
+        # if "errors" in response_data:
+        #     return TrackingResult(False)
+        return mps
+
     def download_pod(self, unique_id, new_filename):
         if len(unique_id.split('~')) == 1:
             result = self.track_by_number(unique_id)
